"""
runner interface.
"""

from danoan.llm_assistant.common import model
from danoan.llm_assistant.common.logging_config import setup_logging

from danoan.llm_assistant.runner.core import exception

from langchain_openai import ChatOpenAI
from langchain_core.prompts import ChatPromptTemplate

import logging
from typing import List

setup_logging()
logger = logging.getLogger(__name__)

################################################################
# LLM Assistant Instance
################################################################


def _singleton(cls):
    instances_dict = {}

    def inner(*args, **kwargs):
        if cls not in instances_dict:
            instances_dict[cls] = cls(*args, **kwargs)
        return instances_dict[cls]

    return inner


@_singleton
class LLMAssistant:
    def __init__(self):
        self._config = None

    def setup(self, config: model.RunnerConfiguration):
        self._config = config
        if self._config.use_cache:
            from langchain.globals import set_llm_cache
            from langchain_community.cache import SQLiteCache

            set_llm_cache(SQLiteCache(database_path=self._config.cache_path))

    @property
    def config(self) -> model.LLMAssistantConfiguration:
        if not self._config:
            raise exception.LLMAssistantNotConfiguredError
        return self._config


################################################################
# Commands
################################################################


def get_prompt_input_variables(
    prompt_configuration: model.PromptConfiguration,
) -> List[str]:
    """
    Get input variables of a prompt.
    """
    prompt = ChatPromptTemplate.from_messages(
        [
            ("system", prompt_configuration.system_prompt),
            ("user", prompt_configuration.user_prompt),
        ]
    )

    return prompt.input_variables


def custom(
    prompt_configuration: model.PromptConfiguration,
    **prompt_variables,
):
    """
    Execute a custom prompt.
    """
    instance = LLMAssistant()

    model = prompt_configuration.model
    if not model:
        model = instance.config.model

    llm = ChatOpenAI(
        api_key=instance.config.openai_key,
        model=model,
    )

    prompt = ChatPromptTemplate.from_messages(
        [
            ("system", prompt_configuration.system_prompt),
            ("user", prompt_configuration.user_prompt),
        ]
    )

    logger.debug(f"Running prompt\n{prompt}")
    chain = prompt | llm
    response = chain.invoke(prompt_variables)

    return response



"""
Data models shared by prompt and runner modules.
"""

from dataclasses import asdict, dataclass
from pathlib import Path
from typing import Any, Dict, Optional

#################################
# Configuration data
#################################


@dataclass
class PromptRepositoryConfiguration:
    git_user: str
    prompt_collection_folder: Path
    versioning: Optional[Dict[str, str]] = None

    def __post_init__(self):
        self.prompt_collection_folder = Path(self.prompt_collection_folder)

    def __asdict__(self) -> Dict[str, str]:
        d = asdict(self)
        d["prompt_collection_folder"] = str(self.prompt_collection_folder)

        return d

    def __str__(self):
        return (
            f"git_user: {self.git_user}\n"
            f"prompt_collection_folder: {str(self.prompt_collection_folder)}\n"
            f"versioning: {self.versioning}\n"
        )


@dataclass
class RunnerConfiguration:
    openai_key: Optional[str] = None
    model: Optional[str] = None
    use_cache: bool = False
    cache_path: Optional[Path] = None

    def __post_init__(self):
        self.cache_path = Path(self.cache_path) if self.cache_path else None

    def __asdict__(self):
        d = asdict(self)
        if self.cache_path:
            d["cache_path"] = str(self.cache_path)

        return d

    def __str__(self):
        return (
            f"openai_key: <<HIDDEN>>\n"
            f"model: {self.model}\n"
            f"use_cache: {self.use_cache}\n"
            f"cache_path: {str(self.cache_path)}\n"
        )


@dataclass
class LLMAssistantConfiguration:
    runner: Optional[RunnerConfiguration] = None
    prompt: Optional[PromptRepositoryConfiguration] = None

    def __str__(self):
        return (
            f"Runner configuration\n\n{self.runner}\n\n"
            f"Prompt repository configuration\n\n{self.prompt}"
        )

    def __asdict__(self):
        return {
            "runner": self.runner.__asdict__()
            if type(self.runner) is RunnerConfiguration
            else self.runner,
            "prompt": self.prompt.__asdict__()
            if type(self.prompt) is PromptRepositoryConfiguration
            else self.prompt,
        }

    @classmethod
    def from_dict(
        cls,
        runner: Optional[Dict[str, Any]] = None,
        prompt: Optional[Dict[str, Any]] = None,
    ):
        inst_runner = None
        inst_prompt = None
        if runner:
            inst_runner = RunnerConfiguration(**runner)
        if prompt:
            inst_prompt = PromptRepositoryConfiguration(**prompt)

        return cls(inst_runner, inst_prompt)


#################################
# Application entities
#################################


@dataclass
class PromptConfiguration:
    name: str
    system_prompt: str
    user_prompt: str
    model: Optional[str] = None



from danoan.llm_assistant.runner.cli.commands.run import run_parser as run
from danoan.llm_assistant.runner.cli.commands.session import session_parser as session
from danoan.llm_assistant.runner.cli.commands.setup import setup

import argparse


def main():
    parser = argparse.ArgumentParser(
        "llm-assistant", description="Collection of LLM applications"
    )
    subcommand_action = parser.add_subparsers()

    list_of_commands = [setup, run, session]
    for command in list_of_commands:
        command.extend_parser(subcommand_action)

    args = parser.parse_args()
    if "func" in args:
        args.func(**vars(args))
    elif "subommand_help" in args:
        args.subommand_help()
    else:
        parser.print_help()


if __name__ == "__main__":
    main()



import argparse


def __session__(*args, **kwargs):
    """
    Chat interface to interact with pre-registered prompts.

    To start a chat session the prompt must need to be stored
    in the prompt repository folder, configurable from the
    application settings.

    Ctrl+B: restart the session
    Ctrl+Q: exit the session
    """

    from danoan.llm_assistant.runner.cli.commands.session import session as S

    S.start_session()


def extend_parser(subparser_action=None):
    command_name = "session"
    description = __session__.__doc__
    help = description.split(".")[0] if description else ""

    if subparser_action:
        parser = subparser_action.add_parser(
            command_name,
            help=help,
            description=description,
            formatter_class=argparse.RawDescriptionHelpFormatter,
        )
    else:
        parser = argparse.ArgumentParser(
            command_name,
            description=description,
            formatter_class=argparse.RawDescriptionHelpFormatter,
        )

    parser.set_defaults(func=__session__, subcommand_help=parser.print_help)

    return parser



from pathlib import Path
from typing import Any, List

from rich import print
from rich.columns import Columns
from rich.console import Console
from rich.panel import Panel
from rich.text import Text

from danoan.llm_assistant.common import config
from danoan.llm_assistant.prompt.core import api, utils

###################################
# GUI Primitives
###################################


_console = Console()


def _make_numbered_list(entries: List[Any], padding=(0, 10), **kwargs) -> Columns:
    list_elements = [f"{i}. {s}" for i, s in enumerate(entries, 1)]
    return Columns(list_elements, padding=padding, **kwargs)


def _make_numbered_vertical_list(entries: List[Any]) -> Text:
    text = Text()
    for i, e in enumerate(entries, 1):
        text.append(f"{i}. {e}")
        if i != len(entries):
            text.append("\n")
    return text


def print_list(*args, **kwargs):
    _console.print(_make_numbered_list(*args, **kwargs))


def print_vertical_list(entries: List[Any]):
    _console.print(_make_numbered_vertical_list(entries))


def print_panel_list(title, entries: List[Any], **kwargs):
    columns = _make_numbered_list(entries)
    panel = Panel(columns, title=title, **kwargs)
    _console.print(panel)


def print_vertical_panel_list(title, entries: List[Any], **kwargs):
    panel = Panel(_make_numbered_vertical_list(entries), title=title)
    _console.print(panel)


def print_panel_table(title, rows, **kwargs):
    table = Columns(rows, column_first=True)
    panel = Panel(table, title=title, **kwargs)
    _console.print(panel)


###################################
# GUI Common Derivatives
###################################


def print_commits(repository_folder: Path, commits_hashes: List[str]):
    """
    Print a summary of the given commit hashes.

    <Hash> <title> <tags>
    """
    rows = []
    for commit_hash in commits_hashes:
        commit = utils.get_commit(repository_folder, commit_hash)
        sha = Text(commit.hexsha)
        title = Text(commit.message.splitlines()[0])
        tags = ",".join(
            [t.name for t in utils.get_commit_tags(repository_folder, commit_hash)]
        )
        list_tags = Text(f"({tags})")
        rows.append(Columns([sha, title, list_tags]))
    print_panel_table("Previous Commits", rows)


def print_previous_commit(repository_folder: Path, commit_hash: str):
    """
    Print a summary of all commits previous the given one.
    """
    most_recent_tags = utils.get_most_recent_tags_before_commit(
        repository_folder, commit_hash
    )
    if len(most_recent_tags) == 0:
        return
    tag = most_recent_tags[-1]
    hashes = utils.get_all_commits_in_between(
        repository_folder, commit_hash, tag.commit
    )
    print_commits(repository_folder, hashes)


def print_staging_area(repository_folder: Path):
    """
    Print respoitory staging area.
    """
    row = []
    for i, entry in enumerate(utils.get_staging_area(repository_folder), 1):
        row.append(Columns([str(i), entry.a_path, entry.change_type]))

    print_panel_table("Staging Area", row)


def print_side_by_side(
    left_content, right_content, left_title="Left", right_title="Right"
):
    console = Console()

    terminal_width = console.size.width
    column_width = terminal_width // 2 - 1

    row = [
        Panel(left_content, style="orange1", title=left_title, width=column_width),
        Panel(right_content, style="pink3", title=right_title, width=column_width),
    ]

    c = Columns(row, equal=True, expand=True)
    print(c)


###################################
# Configuration file
###################################


def ensure_configuration_file_exists():
    if not config.get_configuration_filepath().exists():
        print(
            f"Configuration file for prompt-evaluator does not exist. It should be located \
            at {config.get_configuration_filepath()}"
        )
        exit(1)


def ensure_prompt_collection_folder_exists():
    prompt_collection_folder = api.get_prompts_folder()

    if not prompt_collection_folder.exists():
        print(
            f"The prompt collection folder {prompt_collection_folder} does not exist. Try to run sync first"
        )
        exit(1)


def ensure_prompt_repository_exists(repository_name: str):
    repository_folder = api.get_prompts_folder() / repository_name

    if not api.is_prompt_repository(repository_folder):
        print(f"The path {repository_folder} does not point to a prompt repository.")
        exit(1)



from danoan.llm_assistant.prompt.core import api
from danoan.llm_assistant.prompt.cli import utils as cli_utils

from typing import List


def __list_prompts__() -> List[str]:
    entries = []
    for tp in api.get_tracked_prompts():
        entry = f"{tp.name}: @{tp.current_tag}"
        entries.append(entry)
    return entries


def list_prompts():
    cli_utils.ensure_prompt_collection_folder_exists()
    cli_utils.print_panel_list("Tracked Prompts", __list_prompts__())



# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

# Add the package source to the sys.path
import sphinx_rtd_theme
from importlib import metadata
import pathlib
import sys

source_folder = (
    pathlib.Path(__file__).parents[1].joinpath("src/danoan").resolve().as_posix()
)
sys.path.insert(0, source_folder)

# Tell sphinx the package version

PACKAGE_VERSION = metadata.version("llm-assistant")
version = release = PACKAGE_VERSION

# Import read the docs theme

project = "llm-assistant"
copyright = "2024, Daniel Martins Antunes"
author = "Daniel Martins Antunes"

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = [
    "sphinx.ext.autodoc",  # Collect docstrings
    "sphinx.ext.autodoc.typehints",  # Use typehints
    "sphinx.ext.autosummary",  # Used to create a short TOC for the reference documentation
    "sphinx.ext.napoleon",  # NumPy and Google docstring format
    "sphinx.ext.viewcode",  # Source code link at function, class, module documentation
    "sphinx_rtd_theme",  # Read the docs theme,
    "myst_parser",  # Markdown flavour. Allow type-hints constructions.
    "sphinxcontrib.mermaid",  # Mermaid diagrams.
]

myst_enable_extensions = ["colon_fence"]

source_suffix = {
    ".rst": "restructuredtext",
    ".md": "markdown",
}

autodoc_typehints = "both"

templates_path = ["_templates"]
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store"]


# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_theme = "sphinx_rtd_theme"
html_static_path = ["_static"]

html_css_files = ["css/custom.css"]



from danoan.llm_assistant.common.model import (
    RunnerConfiguration,
    PromptRepositoryConfiguration,
)
from danoan.llm_assistant.common.config import get_absolute_configuration_path

from copy import deepcopy


def generate_absolute_runner_config(runner_config: RunnerConfiguration):
    """
    Resolve all paths in the runner configuration to absolute paths.

    This requires llm-assistant-config.toml to be defined.
    """
    runner_config_copy = deepcopy(runner_config)
    runner_config_copy.cache_path = get_absolute_configuration_path(
        runner_config_copy.cache_path
    )

    return runner_config_copy


def generate_absolute_prompt_config(prompt_repo_config: PromptRepositoryConfiguration):
    """
    Resolve all paths in the prompt repository configuration to absolute paths.

    This requires llm-assistant-config.toml to be defined.
    """
    prompt_repo_config_copy = deepcopy(prompt_repo_config)
    prompt_repo_config_copy.prompt_collection_folder = get_absolute_configuration_path(
        prompt_repo_config_copy.prompt_collection_folder
    )

    return prompt_repo_config_copy



from danoan.llm_assistant.prompt.cli.commands.versioning import parser as versioning
from danoan.llm_assistant.prompt.cli.commands.evaluate import parser as evaluate

import argparse


def main():
    parser = argparse.ArgumentParser(
        "prompt-evaluator",
        description="Collection of tools to evaluate and versioning prompts",
    )
    subparser_action = parser.add_subparsers()

    list_of_commands = [evaluate, versioning]
    for command in list_of_commands:
        command.extend_parser(subparser_action)

    args = parser.parse_args()
    if "func" in args:
        args.func(**vars(args))
    elif "subcommand_help" in args:
        args.subcommand_help()
    else:
        parser.print_help()


if __name__ == "__main__":
    main()



"""
LLM-assistant configuration API.
"""
from danoan.llm_assistant.common import exception, model
from danoan.llm_assistant.common.logging_config import setup_logging

import logging
import os
from functools import lru_cache
from pathlib import Path
import toml
from typing import Optional

setup_logging()
logger = logging.getLogger(__name__)

########################################
# Configuration files
########################################

LLM_ASSISTANT_ENV_VARIABLE = "LLM_ASSISTANT_CONFIGURATION_FOLDER"
LLM_ASSISTANT_CONFIGURATION_FILENAME = "llm-assistant-config.toml"


@lru_cache
def _get_first_configuration_filepath_within_file_hierarchy(
    base_dir: Path,
) -> Optional[Path]:
    """
    Traverses the parents of the working directory until
    the configuration file is found.
    """
    visited = set()
    folders_to_visit = [base_dir]
    while len(folders_to_visit) > 0:
        cur_folder = folders_to_visit.pop()
        if cur_folder in visited:
            break
        logger.debug(f"Visiting {cur_folder}")
        visited.add(cur_folder)
        folders_to_visit.append(cur_folder.parent)
        for p in cur_folder.iterdir():
            if not p.is_dir():
                if p.name == LLM_ASSISTANT_CONFIGURATION_FILENAME:
                    return p
                continue
    return None


def get_configuration_folder() -> Path:
    """
    Return directory where configuration file is stored.

    First checks if a configuration file exists in the file hierarchy.
    If that is the case, return the directory where the configuration file
    is located.

    If the procedure above does not find a configuration file, return the
    value stored in the environment variable LLM_ASSISTANT_ENV_VARIABLE.

    If the environment variable is not defined, raise an error.

    Raises:
        EnvironmentVariableNotDefinedError: If the LLM_ASSISTANT_ENV_VARIABLE
                                            is not defined and a configuration file
                                            is not found in the file hierarchy
    """
    logger.debug("Start hierarchical search of configuation file")
    config_filepath = _get_first_configuration_filepath_within_file_hierarchy(
        Path(os.getcwd())
    )
    if config_filepath:
        return config_filepath.parent

    logger.debug(
        "Hierarchical search failed. Check environment variable {LLM_ASSISTANT_ENV_VARIABLE}"
    )
    if LLM_ASSISTANT_ENV_VARIABLE in os.environ:
        return Path(os.environ[LLM_ASSISTANT_ENV_VARIABLE]).expanduser()

    raise exception.EnvironmentVariableNotDefinedError()


def get_environment_variable_value() -> Path:
    f"""
    Return the value stored by {LLM_ASSISTANT_ENV_VARIABLE}.

    Raises:
        EnvironmentVariableNotDefinedError: If the LLM_ASSISTANT_ENV_VARIABLE
                                            is not defined and a configuration file
                                            is not found in the file hierarchy
    """
    if LLM_ASSISTANT_ENV_VARIABLE in os.environ:
        return Path(os.environ[LLM_ASSISTANT_ENV_VARIABLE]).expanduser()

    raise exception.EnvironmentVariableNotDefinedError()


def get_configuration_filepath() -> Path:
    """
    Return path to llm-assistant configuration file.
    """
    return get_configuration_folder() / LLM_ASSISTANT_CONFIGURATION_FILENAME


def get_configuration() -> model.LLMAssistantConfiguration:
    """
    Return configuration object.
    """
    config_filepath = get_configuration_filepath()
    if not config_filepath.exists():
        raise exception.ConfigurationFileDoesNotExistError()

    with open(config_filepath, "r") as f:
        return model.LLMAssistantConfiguration.from_dict(**toml.load(f))


def get_prompt_configuration(prompt_name: str) -> model.PromptConfiguration:
    """
    Get prompt configuration object.

    It searches the prompt configuration file within the directory specified
    by runner.prompt_collection_folder setting.

    Raises:
        FileNotFoundError: if prompt configuration file is not found.
    """
    config = get_configuration()
    if config.prompt:
        prompt_config_filepath = get_absolute_configuration_path(
            config.prompt.prompt_collection_folder / prompt_name / "config.toml"
        )
    else:
        raise FileNotFoundError(
            2, "File not found", "prompt collection folder is not specified"
        )

    if not prompt_config_filepath.exists():
        raise FileNotFoundError(2, "File not found", prompt_config_filepath)
    with open(prompt_config_filepath) as f:
        return model.PromptConfiguration(**toml.load(f))


def get_absolute_configuration_path(path: Path):
    """
    Get absolute path of a configuration parameter.

    Paths in the configuration file are given relative to the location of
    the configuration file. This function resolves to its absolute path.
    """
    if path.is_absolute():
        return path
    else:
        return get_configuration_folder() / path



from pathlib import Path
from typing import Any, Tuple

import toml

from danoan.llm_assistant.common import config, exception, model
from danoan.llm_assistant.runner.core import api


def ensure_configuration_file_exists(logger):
    try:
        config.get_configuration()
    except exception.ConfigurationFileDoesNotExistError:
        logger.error(
            f"The file {config.get_configuration_filepath()} was not found. You can create one by calling llm-assistant setup init"
        )
        exit(1)


def ensure_prompt_exists(prompt_name: str, logger):
    llma_config = config.get_configuration()

    if not llma_config.prompt:
        logger.error("Prompt settings are not specified")
        exit(1)

    prompt_config_filepath = config.get_absolute_configuration_path(
        llma_config.prompt.prompt_collection_folder / prompt_name / "config.toml"
    )
    if not prompt_config_filepath.exists():
        logger.error(
            f"Could not find the configuration file for prompt {prompt_name}. It should be located at {prompt_config_filepath}"
        )
        exit(1)


def normalize_name(name: str) -> str:
    return name.lower().replace(" ", "-")


def value_or_default(data_dict, key, default: Any):
    if key in data_dict:
        return data_dict[key]
    else:
        return default


def is_a_prompt_config_file(
    filepath: Path,
) -> Tuple[bool, model.PromptConfiguration]:
    try:
        o = toml.load(filepath)
        prompt_config = model.PromptConfiguration(**o)
        return True, prompt_config
    except toml.TomlDecodeError as ex:
        raise ex
    except ValueError:
        return False, None



from danoan.llm_assistant.common.logging_config import setup_logging
from danoan.llm_assistant.common import config
from danoan.llm_assistant.common import model

import logging
import os
from pathlib import Path
import toml

setup_logging()
logger = logging.getLogger(__name__)


def __create_configuration_model__():
    runner_config = model.RunnerConfiguration(
        "openai-key", "openai-model", True, "path to the cache file"
    )
    repo_config = model.PromptRepositoryConfiguration(
        "git-namespace-prompt-repository",
        "local folder where prompts are stored",
        {"my-prompt": "1.0.0"},
    )
    return model.LLMAssistantConfiguration(runner_config, repo_config)


def init(reset: bool = False, use_env_var: bool = False):
    if use_env_var:
        config_folder = config.get_configuration_folder()
        if not config_folder.exists():
            config_folder.mkdir(parents=True, exist_ok=True)

        config_path = config.get_configuration_filepath()
    else:
        config_path = Path(os.getcwd()) / config.LLM_ASSISTANT_CONFIGURATION_FILENAME

    if config_path.exists() and not reset:
        logger.info(f"The configuration file: {config_path} exists already.")

    if config_path.exists() and reset:
        logger.info(f"Reseting the current configuration file: {config_path}.")

    if not config_path.exists() or reset:
        llma_config = __create_configuration_model__()
        with open(config_path, "w") as f:
            toml.dump(llma_config.__asdict__(), f)



import argparse


def __list__(*args, **kwargs):
    """
    List all tracked prompts.
    """
    from danoan.llm_assistant.prompt.cli.commands.versioning.commands.list import (
        action as A,
    )

    A.list_prompts()


def extend_parser(subparser_action=None):
    command_name = "list"
    description = __list__.__doc__
    help = description.split(".")[0] if description else ""

    if subparser_action:
        parser = subparser_action.add_parser(
            command_name,
            help=help,
            description=description,
            formatter_class=argparse.RawDescriptionHelpFormatter,
        )
    else:
        parser = argparse.ArgumentParser(
            command_name,
            description=description,
            formatter_class=argparse.RawDescriptionHelpFormatter,
        )

    parser.set_defaults(func=__list__, subcommand_help=parser.print_help)

    return parser



from danoan.llm_assistant.prompt.cli.commands.versioning.commands.list import (
    parser as list_prompts,
)
from danoan.llm_assistant.prompt.cli.commands.versioning.commands.push import (
    parser as push,
)
from danoan.llm_assistant.prompt.cli.commands.versioning.commands.sync import (
    parser as sync,
)

import argparse


def extend_parser(subparser_action=None):
    command_name = "versioning"
    description = "Collection of tools for prompt versioning"
    help = description.split(".")[0] if description else ""

    if subparser_action:
        parser = subparser_action.add_parser(
            command_name,
            help=help,
            description=description,
            formatter_class=argparse.RawDescriptionHelpFormatter,
        )
    else:
        parser = argparse.ArgumentParser(
            command_name,
            description=description,
            formatter_class=argparse.RawDescriptionHelpFormatter,
        )

    subparser_action = parser.add_subparsers()
    list_of_commands = [list_prompts, push, sync]
    for command in list_of_commands:
        command.extend_parser(subparser_action)

    parser.set_defaults(subcommand_help=parser.print_help)

    return parser



from danoan.llm_assistant.common.logging_config import setup_logging
from danoan.llm_assistant.common import config, utils
from danoan.llm_assistant.runner.cli import utils as cli_utils
from danoan.llm_assistant.runner.cli.commands.session import session_core as core
from danoan.llm_assistant.runner.cli.commands.session.cli_drawer import CLIDrawer
from danoan.llm_assistant.runner.cli.commands.session.task_runner import (
    Task,
    TaskInstruction,
    TaskRunner,
)
from danoan.llm_assistant.runner.core import api

import fcntl
import logging
import os
from pathlib import Path
import selectors
import sys

from pynput import keyboard
from pynput.keyboard import Controller, Key

from rich.columns import Columns
from rich.console import Console as RichConsole
from rich.panel import Panel
from rich.text import Text


setup_logging()
logger = logging.getLogger(__name__)


def create_non_blocking_input():
    # Set up the selector
    sel = selectors.DefaultSelector()
    sel.register(sys.stdin, selectors.EVENT_READ)

    # Make stdin non-blocking
    fd = sys.stdin.fileno()
    fl = fcntl.fcntl(fd, fcntl.F_GETFL)
    fcntl.fcntl(fd, fcntl.F_SETFL, fl | os.O_NONBLOCK)

    def input_reader():
        events = sel.select(timeout=0.1)
        for key, mask in events:
            if key.fileobj is sys.stdin:
                try:
                    user_input = sys.stdin.readline().split("\n")[0]
                    return user_input
                except IOError:
                    pass
        return None

    return input_reader


non_blocking_input = create_non_blocking_input()


#############################################
# Rich Command Line
#############################################


class _RichCLIDrawer(CLIDrawer):
    def __init__(self):
        self._console = RichConsole()

    def print_error(self, **kwargs):
        message = kwargs["message"]
        panel = Panel(message, style="red")

        self._console.print(panel)

    def print_panel(self, **kwargs):
        message = kwargs["message"]
        title = cli_utils.value_or_default(kwargs, "title", None)
        color = cli_utils.value_or_default(kwargs, "color", "")

        panel = Panel(message, style=color, title=title)
        self._console.print(panel)

    def print_list(self, **kwargs):
        list_elements = kwargs["list_elements"]
        numbered = cli_utils.value_or_default(kwargs, "numbered", True)

        if numbered:
            list_elements = [f"{i}. {s}" for i, s in enumerate(list_elements, 1)]

        columns = Columns(list_elements)
        self._console.print(columns)
        self._console.print("")

    def prompt(self, **kwargs):
        message = kwargs["message"]
        task = kwargs["task"]

        text = Text(message)
        text.stylize("bold")
        self._console.print(text, end="")

        value = None
        with task.get_event_notifier(Task.Event.Stop) as event_notifier:
            while not event_notifier.is_set() and value is None:
                value = non_blocking_input()

        if value is None:
            keyboard = Controller()
            keyboard.press(Key.enter)
            keyboard.release(Key.enter)
            _ = non_blocking_input()
            raise core.PromptInterruptedError()
        else:
            # Remove session commands from  input
            value = value.replace(f"\x02", "")  # Ctrl+B
            value = value.replace(f"\x11", "")  # Ctrl+Q

        return value


#############################################
# Keyboard Event Listener
#############################################


class KeyboardListener:
    def __init__(self):
        self._events = {}

    def register(self, key_sequence, callback):
        cur_node = self._events
        for key in key_sequence:
            if key not in cur_node:
                cur_node[key] = {}
            cur_node = cur_node[key]

        cur_node["callback"] = callback

    def _on_press(self):
        cur_node = self._events

        def inner(key):
            nonlocal cur_node

            effective_key = None
            if key in cur_node:
                effective_key = key
            elif isinstance(key, keyboard.KeyCode) and key.char in cur_node:
                effective_key = key.char

            if effective_key:
                cur_node = cur_node[effective_key]
            else:
                cur_node = self._events

            if "callback" in cur_node:
                cur_node["callback"]()
                cur_node = self._events

        return inner

    def start(self):
        listener = keyboard.Listener(on_press=self._on_press())
        listener.start()


#############################################
# Argument Parser
#############################################


def start_session():
    llma_config = config.get_configuration()

    if not llma_config.runner:
        logger.error("Runner is not configured. Please use the setup command")
        exit(1)

    if not llma_config.prompt:
        logger.error(
            "Prompt repository path is not configured. Please use the setup command"
        )
        exit(1)

    api.LLMAssistant().setup(utils.generate_absolute_runner_config(llma_config.runner))
    prompt_repository = config.get_absolute_configuration_path(
        Path(llma_config.prompt.prompt_collection_folder)
    )
    if not prompt_repository.exists():
        logger.error("Pront repository does not exist. Please use the setup command")
        exit(1)
    cliDrawer = _RichCLIDrawer()
    tr = TaskRunner()
    core.register_tasks(
        cliDrawer,
        prompt_repository,
        tr.register,
    )

    def restart_runner():
        tr.clear()
        tr.add(TaskInstruction(core.TaskName.NoPromptSelected, None))
        tr.next()

    def exit_runner():
        tr.clear()
        tr.stop()
        tr.next()

    # Create a listener
    KL = KeyboardListener()
    KL.register([Key.ctrl_l, "b"], restart_runner)
    KL.register([Key.ctrl_l, "q"], exit_runner)
    KL.start()

    tr.run()



"""
prompt-manager interface.
"""

from danoan.llm_assistant.common import config
from danoan.llm_assistant.common.logging_config import setup_logging
from danoan.llm_assistant.common.model import PromptRepositoryConfiguration
from danoan.llm_assistant.prompt.core import model, utils

import copy
from dataclasses import asdict, dataclass
from enum import Enum
import git
import logging
from pathlib import Path
import re
import toml
from typing import Generator, List, Literal, Optional

setup_logging()
logger = logging.getLogger(__name__)


def get_prompts_folder() -> Path:
    """
    Return path to the folder where all tracked prompts are located.
    """
    llma_config = config.get_configuration()
    return config.get_absolute_configuration_path(
        llma_config.prompt.prompt_collection_folder
    )


def get_prompt_configuration_filepath(prompt_name: str) -> Path:
    """
    Return path to prompt configuration file.
    """
    return get_prompts_folder() / prompt_name / "config.toml"


def is_prompt_repository(path: Path) -> bool:
    """
    Check if a path points to a prompt repository.

    A prompt repository must have a `config.toml`
    that must have the following keys.

    - user_prompt
    - system_prompt

    """
    configuration_prompt_path = path / "config.toml"
    if not configuration_prompt_path.exists():
        return False

    logger.debug(f"Loading: {configuration_prompt_path}")
    obj = toml.load(configuration_prompt_path)
    mandatory_keys = ["user_prompt", "system_prompt"]
    for key in mandatory_keys:
        if key not in obj:
            logger.debug(f"Missing mandatory key {key}")
            return False
    return True


def get_tracked_prompt(repository_name: str) -> model.TrackedPrompt:
    """
    Return a TrackedPrompt from the prompt repository.

    Raises:
        FileNotFoundError: if prompt folder does not exist.
        AttributeError: if head points to an invalid object or referece.
        git.exc.InvalidGitRepositoryError: if prompt folder is not a valid git repository.
    """
    repository_folder = get_prompts_folder() / repository_name
    if not repository_folder.exists():
        raise FileNotFoundError()
    repository = git.Repo(repository_folder)
    current_tag = None
    if repository.head.is_valid():
        current_commit = repository.head.commit
        for tag in repository.tags:
            if tag.commit == current_commit:
                current_tag = tag.name
    branches = [b for b in repository.branches if b.name not in ["master"]]
    logger.debug(f"Branches of {repository_name}: {branches}")
    return model.TrackedPrompt(
        repository_name, repository_folder, current_tag, branches
    )


def get_tracked_prompts() -> Generator[model.TrackedPrompt, None, None]:
    """
    Return a list of all prompts tracked by the tool.

    A prompt is considered tracked if it is a prompt repository located
    at `get_prompts_folder`.

    A prompt repository is a folder with a `config.toml` file that
    can be loaded as a PromptConfiguration.
    """
    prompts_folder = get_prompts_folder()
    for x in prompts_folder.iterdir():
        if not is_prompt_repository(x):
            continue
        yield get_tracked_prompt(x.name)


def get_prompt_test_regression_filepath(prompt_name: str) -> Path:
    """
    Get regression test file.
    """
    return (
        get_prompts_folder() / prompt_name / "tests" / "regression" / "regression.json"
    )


def __resolve_version__(prompt_name: str, version: str):
    tp = get_tracked_prompt(prompt_name)
    if version.strip() == "" or version.lower() == "last":
        all_versions = get_prompt_versions(tp.repository_path)
        if len(all_versions) == 0:
            return version
        return all_versions[-1]
    return version


def sync(repo_config: PromptRepositoryConfiguration, progress_callback=None):
    """
    Read prompt repository configuration file and sync local folder
    with the specified prompt version.
    """

    class Events(Enum):
        SYNC_CONFIG = "sync_config"
        FETCH = "fetch"
        CHECKOUT = "checkout"
        SYNCED = "synced"
        SYNC_LOCAL_FOLDER = "sync_prompt_collection_folder"
        NOT_TRACKED = "not_tracked"
        NOT_PROMPT_REPOSITORY = "not_prompt_repository"
        GIT = "git"
        ITEM = "item"
        BEGIN = "begin"
        END = "end"

    @dataclass
    class SyncItem:
        event: Literal[
            Events.SYNC_CONFIG,
            Events.FETCH,
            Events.CHECKOUT,
            Events.SYNCED,
            Events.SYNC_LOCAL_FOLDER,
            Events.NOT_TRACKED,
            Events.NOT_PROMPT_REPOSITORY,
            Events.GIT,
            Events.ITEM,
            Events.BEGIN,
            Events.END,
        ]
        name: Optional[str] = None
        value: Optional[str] = None

    def _progress_callback(sync_item: SyncItem):
        if progress_callback:
            d = asdict(sync_item)
            d["event"] = sync_item.event.value
            progress_callback(**d)
        else:
            return

    def _git_progress_callback(op_code, cur_count, max_count=None, message=""):
        _progress_callback(SyncItem(Events.GIT))
        _progress_callback(SyncItem(Events.BEGIN))
        _progress_callback(SyncItem(Events.ITEM, "op_code", op_code))
        _progress_callback(SyncItem(Events.ITEM, "cur_count", cur_count))
        _progress_callback(SyncItem(Events.ITEM, "max_count", max_count))
        _progress_callback(SyncItem(Events.ITEM, "message", message))
        _progress_callback(SyncItem(Events.END))

    # Sync prompt repository configuration
    logger.debug("Align prompt collection folder with configuration file")
    _progress_callback(SyncItem(Events.SYNC_CONFIG))
    for prompt_name, version in repo_config.versioning.items():
        logger.debug(f"Start syncing of {prompt_name} {version}")

        _progress_callback(SyncItem(Events.SYNC_CONFIG, "prompt_name", prompt_name))
        _progress_callback(SyncItem(Events.SYNC_CONFIG, "version", version))
        prompt_folder = get_prompts_folder() / prompt_name
        prompt_repo_url = f"https://github.com/{repo_config.git_user}/{prompt_name}.git"
        repo = None
        if not prompt_folder.exists():
            _progress_callback(SyncItem(Events.FETCH, "prompt", prompt_repo_url))
            repo = git.Repo.clone_from(
                prompt_repo_url, prompt_folder, progress=_git_progress_callback
            )
        else:
            repo = git.Repo(prompt_folder)
        version = __resolve_version__(prompt_name, version)
        repo.remote().fetch(tags=True)
        _progress_callback(SyncItem(Events.CHECKOUT, "version", version))
        repo.git.checkout(f"tags/v{version}")

    # Sync prompt repository local folder
    logger.debug("Align configuration file with prompt collection folder")
    updated_repo_config = copy.deepcopy(repo_config)
    _progress_callback(SyncItem(Events.SYNC_LOCAL_FOLDER))

    for prompt_folder in get_prompts_folder().iterdir():
        _progress_callback(SyncItem(Events.SYNC_LOCAL_FOLDER, "folder", prompt_folder))
        if is_prompt_repository(prompt_folder):
            prompt_name = prompt_folder.stem
            if prompt_name not in repo_config.versioning.keys():
                tp = get_tracked_prompt(prompt_name)
                updated_repo_config.versioning[prompt_name] = tp.current_tag
                _progress_callback(
                    SyncItem(Events.NOT_TRACKED, "version", tp.current_tag)
                )
            else:
                _progress_callback(SyncItem(Events.SYNCED))
        else:
            _progress_callback(SyncItem(Events.NOT_PROMPT_REPOSITORY))

    llma_config = config.get_configuration()
    llma_config.prompt = updated_repo_config

    with open(config.get_configuration_filepath(), "w") as f:
        toml.dump(llma_config.__asdict__(), f)


###################################
# Prompt Versioning
###################################


def get_prompt_versions(repository_path: Path) -> List[str]:
    """
    List all the versions available in a local prompt repository.
    """
    l = sorted([model.PromptVersion(v) for v in utils.get_versions(repository_path)])
    return [str(v) for v in l]


def get_most_recent_version_before_commit(
    repository_folder: Path, commit_hash: str
) -> Optional[str]:
    """
    Find the most recent version prior to a given commit hash.
    """
    tags = utils.get_most_recent_tags_before_commit(repository_folder, commit_hash)
    for t in tags:
        if t.name[0] != "v":
            continue

        return t.name[1:]
    return None


def get_current_version(repository_folder: Path) -> Optional[str]:
    """
    Get the most recent version prior to HEAD.
    """
    repo = git.Repo(repository_folder)
    current_tags = utils.get_commit_tags(repository_folder, repo.commit())

    no_version_at_all = (
        len(utils.get_most_recent_tags_before_commit(repository_folder, repo.commit()))
        == 0
    )

    if no_version_at_all:
        return None

    current_commit_is_not_versioned = len(current_tags) == 0 and not no_version_at_all
    most_recent_version = None
    if current_commit_is_not_versioned:
        most_recent_version = get_most_recent_version_before_commit(
            repository_folder, repo.commit()
        )
    else:
        most_recent_version = str(current_tags[0])[1:]

    return most_recent_version


def suggest_next_version(
    repository_path: Path, current_version: str, cn: model.ChangeNature
) -> str:
    """
    Suggest a new version based on the type of changes.

    If PromptTweak, increment the fix unit.
    If InterfaceUpdate, increment the minor unit.
    If ScopeChange, increment the major unit.
    """
    sorted_tags = get_prompt_versions(repository_path)

    vi = model.PromptVersion(current_version, model.PromptVersion.IncrementerType.Fix)
    if cn == model.ChangeNature.PromptTweak:
        vi = model.PromptVersion(
            current_version, model.PromptVersion.IncrementerType.Fix
        )
    elif cn == model.ChangeNature.InterfaceUpdate:
        vi = model.PromptVersion(
            current_version, model.PromptVersion.IncrementerType.Minor
        )
    elif cn == model.ChangeNature.ScopeChange:
        vi = model.PromptVersion(
            current_version, model.PromptVersion.IncrementerType.Major
        )

    while str(vi) in set(sorted_tags):
        vi += 1

    return str(vi)


def update_changelog(changelog_str: str) -> str:
    """
    Rewrites the changelog document such that versions are listed in
    numerical order and divided by major versions.
    """

    def get_next_header(s: str, start: int):
        return s.find("#", start)

    def get_next_version_header(s: str, start: int):
        return s.find("##", start)

    p_version_part = r"\d+[\.xyz]?"
    p = f"({p_version_part}{p_version_part}{p_version_part})"

    content_versions = []
    version_set = set()
    start = 0
    while start != -1:
        match_start = get_next_version_header(changelog_str, start)
        if match_start == -1:
            break
        m = re.search(p, changelog_str[match_start:])

        if not m:
            break

        content_start = match_start + m.span()[1]
        content_end = get_next_header(changelog_str, content_start)
        content = changelog_str[content_start:content_end]
        start = content_end

        version_name = m.group()
        if version_name in version_set:
            continue

        content_versions.append((model.PromptVersion(version_name), content))
        version_set.add(version_name)

    sorted_items = sorted(content_versions, key=lambda x: x[0])

    current_major_title = None
    update_title = False
    rewriten_document = ""
    for item in sorted_items:
        version, content = item

        if current_major_title != version.major:
            current_major_title = version.major
            update_title = True

        if update_title:
            rewriten_document += f"# V{current_major_title}\n\n"
            update_title = False

        rewriten_document += f"## {version}"
        rewriten_document += content + "\n\n"

    return rewriten_document



import argparse


def __regenerate_tests__(prompt_name: str, *args, **kwargs):
    """
    Run all tests and replace stored results.
    """
    from danoan.llm_assistant.prompt.cli.commands.evaluate.commands.regenerate_tests import (
        action as A,
    )

    A.regenerate_tests(prompt_name)


def extend_parser(subparser_action=None):
    command_name = "regenerate-tests"
    description = __regenerate_tests__.__doc__
    help = description.split(".")[0] if description else ""

    if subparser_action:
        parser = subparser_action.add_parser(
            command_name,
            help=help,
            description=description,
            formatter_class=argparse.RawDescriptionHelpFormatter,
        )
    else:
        parser = argparse.ArgumentParser(
            command_name,
            description=description,
            formatter_class=argparse.RawDescriptionHelpFormatter,
        )

    parser.add_argument("prompt_name", type=str, help="Prompt name")
    parser.set_defaults(func=__regenerate_tests__, subcommand_help=parser.print_help)

    return parser



from danoan.llm_assistant.common.logging_config import setup_logging
from danoan.llm_assistant.common.model import PromptConfiguration, RunnerConfiguration
from danoan.llm_assistant.common import config
from danoan.llm_assistant.common import utils

from danoan.llm_assistant.prompt.cli import utils as cli_utils
from danoan.llm_assistant.prompt.core import api
from danoan.llm_assistant.runner.core import api as llma

import git
import json
import logging
import sys
import toml
from tempfile import TemporaryDirectory

setup_logging()
logger = logging.getLogger(__name__)


def __load_prompt_configuration_version__(
    prompt_name: str, version: str
) -> PromptConfiguration:
    tp = api.get_tracked_prompt(prompt_name)
    repo = git.Repo(tp.repository_path)

    if version == "current":
        config_toml = repo.git.show(f"HEAD:config.toml")
    else:
        repo.git.checkout(version)
        config_toml = repo.git.show(f"{version}:config.toml")

    return PromptConfiguration(**toml.loads(config_toml))


def __run_tests__(
    runner_configuration: RunnerConfiguration,
    prompt_name: str,
    base_version: str,
    compare_version: str,
):
    base_prompt_config = __load_prompt_configuration_version__(
        prompt_name, base_version
    )
    compare_prompt_config = __load_prompt_configuration_version__(
        prompt_name, compare_version
    )

    test_model = api.get_prompt_test_regression_filepath(prompt_name)
    test_run = test_model.parent / runner_configuration.model / test_model.name

    with open(test_run, "r") as fi:
        input_obj = json.load(fi)

        for entry in input_obj:
            base_response = llma.custom(base_prompt_config, **entry["input"])
            compare_response = llma.custom(compare_prompt_config, **entry["input"])

            try:
                base_response_obj = json.loads(base_response.content)
            except:
                base_response_obj = base_response.content

            try:
                compare_response_obj = json.loads(compare_response.content)
            except:
                compare_response_obj = compare_response.content

            base_response_str = json.dumps(
                base_response_obj, ensure_ascii=False, indent=2
            )
            compare_response_str = json.dumps(
                compare_response_obj, ensure_ascii=False, indent=2
            )

            yield (base_response_str, compare_response_str)


def run_tests(
    prompt_name: str, use_cache: bool, base_version: str, compare_version: str
):
    runner_configuration = config.get_configuration().runner
    runner_configuration.use_cache = use_cache
    llma.LLMAssistant().setup(
        utils.generate_absolute_runner_config(runner_configuration)
    )

    try:
        for base_response_str, compare_response_str in __run_tests__(
            runner_configuration, prompt_name, base_version, compare_version
        ):
            cli_utils.print_side_by_side(
                base_response_str,
                compare_response_str,
                left_title="Base",
                right_title="Compare",
            )
    except FileNotFoundError as ex:
        print(ex)
        exit(1)



import argparse

from typing import Optional


def __run_tests__(
    prompt_name: str,
    no_cache: bool = False,
    base_version: Optional[str] = None,
    compare_version: Optional[str] = None,
    *args,
    **kwargs,
):
    """
    Run tests for a prompt and generate a report.
    """
    from danoan.llm_assistant.prompt.cli.commands.evaluate.commands.run_tests import (
        action as A,
    )

    use_cache = not no_cache
    if not base_version:
        base_version = "current"
    if not compare_version:
        compare_version = "current"

    A.run_tests(prompt_name, use_cache, base_version, compare_version)


def extend_parser(subparser_action=None):
    command_name = "run-tests"
    description = __run_tests__.__doc__
    help = description.split(".")[0] if description else ""

    if subparser_action:
        parser = subparser_action.add_parser(
            command_name,
            help=help,
            description=description,
            formatter_class=argparse.RawDescriptionHelpFormatter,
        )
    else:
        parser = argparse.ArgumentParser(
            command_name,
            description=description,
            formatter_class=argparse.RawDescriptionHelpFormatter,
        )

    parser.add_argument("prompt_name", type=str)
    parser.add_argument("--no-cache", action="store_true", help="Do not use the cache")
    parser.add_argument(
        "--base-version", "-b", help="Prompt version to be used as base"
    )
    parser.add_argument(
        "--compare-version", "-c", help="Prompt version to be compared with"
    )
    parser.set_defaults(func=__run_tests__, subcommand_help=parser.print_help)

    return parser



from danoan.llm_assistant.common.logging_config import setup_logging
from danoan.llm_assistant.common import config
from danoan.llm_assistant.common import utils
from danoan.llm_assistant.prompt.core import api
from danoan.llm_assistant.runner.core import api as runner
from danoan.llm_assistant.runner.core import api as llma

from copy import deepcopy
import json
import logging
import sys
from typing import Any

setup_logging()
logger = logging.getLogger(__name__)


def __regenerate_tests__(prompt_name: str) -> Any:
    runner_configuration = config.get_configuration().runner
    runner.LLMAssistant().setup(
        utils.generate_absolute_runner_config(runner_configuration)
    )
    prompt_config = config.get_prompt_configuration(prompt_name)

    test_model = api.get_prompt_test_regression_filepath(prompt_name)
    test_run_fp = test_model.parent / runner_configuration.model / test_model.name

    test_run_fp.parent.mkdir(parents=True, exist_ok=True)

    output_obj = []
    with open(test_model, "r") as fi:
        input_obj = json.load(fi)

        for entry in input_obj:
            response = llma.custom(prompt_config, **entry["input"])

            try:
                response_obj = json.loads(response.content)
            except:
                response_obj = response.content

            output_entry = deepcopy(entry)
            output_entry["output"] = response_obj
            output_obj.append(output_entry)

    with open(test_run_fp, "w") as fo:
        json.dump(output_obj, fo, ensure_ascii=False, indent=2)

    return output_obj


def regenerate_tests(prompt_name: str):
    json.dump(
        __regenerate_tests__(prompt_name), sys.stdout, ensure_ascii=False, indent=2
    )



from danoan.llm_assistant.common.logging_config import setup_logging
from danoan.llm_assistant.common import config, exception
from danoan.llm_assistant.runner.cli.commands.setup.setup_commands import (
    init_parser as init,
)

import argparse
import logging

setup_logging()
logger = logging.getLogger(__name__)


def __setup__(*args, **kwargs):
    try:
        print(
            f"The environment variable {config.LLM_ASSISTANT_ENV_VARIABLE}"
            f" is set to: {config.get_environment_variable_value()}"
        )
    except exception.EnvironmentVariableNotDefinedError:
        logger.error(
            f"The environment variable: {config.LLM_ASSISTANT_ENV_VARIABLE} is not set"
        )
        exit(1)

    print(
        f"The configuration file being used is located at: {config.get_configuration_filepath()}\n"
    )
    print(config.get_configuration())


def extend_parser(subparser_action=None):
    command_name = "setup"
    description = "Configure LLM assistant."
    help = description.split(".")[0] if description else ""

    if subparser_action:
        parser = subparser_action.add_parser(
            command_name,
            help=help,
            description=description,
            formatter_class=argparse.RawDescriptionHelpFormatter,
        )
    else:
        parser = argparse.ArgumentParser(
            command_name,
            description=description,
            formatter_class=argparse.RawDescriptionHelpFormatter,
        )

    parser.set_defaults(func=__setup__, subcommand_help=parser.print_help)

    subparser_action = parser.add_subparsers()
    list_of_commands = [init]
    for command in list_of_commands:
        command.extend_parser(subparser_action)

    return parser



from danoan.llm_assistant.common.logging_config import setup_logging
from danoan.llm_assistant.prompt.core import api, model, utils
from danoan.llm_assistant.prompt.cli import utils as cli_utils

import git
import logging
from pathlib import Path
import subprocess
from typing import Callable, List, Optional, Tuple

setup_logging()
logger = logging.getLogger(__name__)

###################################
# Lazy Actions
###################################


def action(fn):
    def wrapper(*args, **kwargs):
        def inner():
            return fn(*args, **kwargs)

        return inner

    return wrapper


@action
def checkout_existing_branch(repository_path: Path, branch_name: str):
    repo = git.Repo(repository_path)
    logger.info(f"Checkout existing branch: {branch_name}")
    repo.git.checkout(branch_name)


@action
def checkout_new_branch(repository_path: Path, branch_name: str):
    repo = git.Repo(repository_path)
    logger.info(f"Creating branch: {branch_name}")
    repo.git.checkout(b=branch_name)


@action
def interactive_add(repository_path: Path):
    subprocess.run(
        [
            "git",
            "add",
            "-i",
        ],
        cwd=repository_path,
    )


@action
def add_file(repository_path: Path, filename: str):
    repo = git.Repo(repository_path)
    repo.git.add(filename)


@action
def commit_message(repository_path: Path, message: str):
    repo = git.Repo(repository_path)
    logger.info(f"Commiting changes: {message}")
    repo.git.commit(m=f"{message}")


@action
def push_branch(repository_path: Path):
    repo = git.Repo(repository_path)
    logger.info(f"Pushing branch: {repo.active_branch}")
    repo.remote().push(repo.active_branch)


@action
def push_tag(repository_path: Path, tag_name: str, message: str):
    repo = git.Repo(repository_path)
    tag_name = f"v{tag_name}"
    logger.info(f"Creating tag: {tag_name}")
    tag = repo.create_tag(tag_name, repo.commit(), message=message)
    repo.remote().push(tag)


@action
def stash_save(repository_path: Path):
    repo = git.Repo(repository_path)
    repo.git.stash("save", "current changes")


@action
def stash_pop(repository_path: Path):
    repo = git.Repo(repository_path)
    repo.git.stash("pop")


@action
def rewrite_readme_file(
    tp: model.TrackedPrompt, version: model.PromptVersion, changes_description: str
):
    readme_file = tp.repository_path / "README.md"

    with open(readme_file) as f:
        readme = f.read()

    readme += f"\n## {version}\n\n{changes_description}\n"
    rewriten_readme = api.update_changelog(readme)

    with open(readme_file, "w") as f:
        f.write(rewriten_readme)


###################################
# Helpers
###################################


def __fetch_remote_data__(tp: model.TrackedPrompt):
    repo = git.Repo(tp.repository_path)
    logger.info("Fetching repository data")
    repo.remote().fetch(tags=True)


def __describe_changes__(tp: model.TrackedPrompt) -> str:
    logger.info("Describe the changes (finish the input by outputing a line with $$)")
    description = ""
    v = input()
    while v != "$$":
        description += f"{v}\n"
        v = input()
    return description


def __update_changelog__(
    tp: model.TrackedPrompt,
    changes_description: str,
    version: model.PromptVersion,
    branch_name: str,
) -> List[Callable[[], None]]:
    git_commands = [
        checkout_existing_branch(tp.repository_path, branch_name),
        rewrite_readme_file(tp, version, changes_description),
        add_file(tp.repository_path, "README.md"),
        commit_message(tp.repository_path, message=f"Update changelog: Add {version}"),
        push_branch(tp.repository_path),
    ]

    return git_commands


def __classify_changes__() -> model.ChangeNature:
    options = [
        "Prompt tweak: Add extra examples, edit intructions to be more precise about some aspect or remove ambiguity.",
        "Interface update: The expected input values or the output format has changed.",
        "Scope change: The prompt become more specialized or more general.",
    ]

    cli_utils.print_vertical_panel_list("Nature of changes", options)
    item_number = input("\nWhat is the nature of the changes? ")
    if item_number == "1":
        return model.ChangeNature.PromptTweak
    elif item_number == "2":
        return model.ChangeNature.InterfaceUpdate
    elif item_number == "3":
        return model.ChangeNature.ScopeChange
    else:
        raise RuntimeError("Invalid option")


def __suggest_version__(tp: model.TrackedPrompt) -> Tuple[str, model.ChangeNature]:
    current_version = api.get_current_version(tp.repository_path)
    if not current_version:
        current_version = "0.0.0"

    sorted_tags = api.get_prompt_versions(tp.repository_path)

    cli_utils.print_panel_list("Current versions", sorted_tags)
    s = input(f"Enter a base version ({current_version}): ")
    if s.strip() != "":
        current_version = s

    cn = __classify_changes__()
    suggested_version = api.suggest_next_version(
        tp.repository_path, current_version, cn
    )

    while True:
        sv_str = input(f"Suggested version ({suggested_version}): ")
        if sv_str.strip() != "":
            suggested_version = model.PromptVersion(sv_str, cn)

        if suggested_version in set(sorted_tags):
            logger.info(
                f"Version {suggested_version} exists already. Please choose another one"
            )
        else:
            break

    return suggested_version, cn


def __commit_changes__(
    tp: model.TrackedPrompt,
    changes_nature: model.ChangeNature,
    changes_description: str,
    version: model.PromptVersion,
) -> List[Callable[[], None]]:
    git_commands = []

    branch_name = f"v{version.major}.{version.minor}"
    repo = git.Repo(tp.repository_path)
    current_branch = repo.active_branch.name
    if current_branch != branch_name:
        logger.error(
            f"You should first move to branch {branch_name} and resolve all eventual conflicts before continue."
        )
        exit(1)

    # Start git add interactive
    git_commands.append(interactive_add(tp.repository_path))

    # Update changelog

    git_commands.extend(
        [
            rewrite_readme_file(tp, version, changes_description),
            add_file(tp.repository_path, "README.md"),
        ]
    )

    # Commit the changes
    message = f"Release version {version}"
    git_commands.append(commit_message(tp.repository_path, message))

    return git_commands


def __create_tag__(
    tp: model.TrackedPrompt,
    changes_nature: model.ChangeNature,
    changes_description: str,
    version: model.PromptVersion,
) -> List[Callable[[], None]]:
    git_commands = []

    branch_name = f"v{version.major}.{version.minor}"
    git_commands.append(checkout_existing_branch(tp.repository_path, branch_name))

    git_commands.append(push_branch(tp.repository_path))

    message = f"Release version {version}\n\n{changes_description}"
    git_commands.append(push_tag(tp.repository_path, str(version), message))
    return git_commands


###################################
# Main
###################################


def push(prompt_name: str, version: Optional[str] = None, *args, **kwargs):
    cli_utils.ensure_configuration_file_exists()
    cli_utils.ensure_prompt_repository_exists(prompt_name)

    tp = api.get_tracked_prompt(prompt_name)
    __fetch_remote_data__(tp)

    changes_description = __describe_changes__(tp)
    suggested_version, changes_nature = __suggest_version__(tp)

    pv = model.PromptVersion(suggested_version)

    git_commands = []
    git_commands.extend(__commit_changes__(tp, changes_nature, changes_description, pv))
    git_commands.extend(__update_changelog__(tp, changes_description, pv, "master"))
    git_commands.extend(__create_tag__(tp, changes_nature, changes_description, pv))

    for command in git_commands:
        command()



import argparse
from typing import Optional


def __push__(prompt_name: str, version: Optional[str] = None, *args, **kwargs):
    """
    Push a new prompt version.

    The version of a prompt is made of three components.
        prompt_id.major.minor

    The prompt_id identifies a prompt variation. For example,
    we may have a prompt to correct text in some language. The
    1.x.x series could be dedicated to very simplistic versions
    of these prompts that only return the corrected version. The
    2.x.x series could be dedicated to versions in which besides
    the correction, the prompt also gives the explanation and
    so on.

    We bump minor every time one of the following situations:
        - Add more examples
        - Any edition on user and system prompts that do
          not modify the input neither output structure.
    We bump the major every time we do one of the following:
        - Update input or output structure.
    """
    from danoan.llm_assistant.prompt.cli.commands.versioning.commands.push import (
        action as A,
    )

    A.push(prompt_name, version)


def extend_parser(subparser_action=None):
    command_name = "push"
    description = __push__.__doc__
    help = description.split(".")[0] if description else ""

    if subparser_action:
        parser = subparser_action.add_parser(
            command_name,
            help=help,
            description=description,
            formatter_class=argparse.RawDescriptionHelpFormatter,
        )
    else:
        parser = argparse.ArgumentParser(
            command_name,
            description=description,
            formatter_class=argparse.RawDescriptionHelpFormatter,
        )

    parser.add_argument("prompt_name", type=str)
    parser.add_argument("--version", type=str)

    parser.set_defaults(func=__push__, subcommand_help=parser.print_help)

    return parser



"""
prompt-manager data models.
"""

from dataclasses import dataclass
from enum import Enum
from functools import total_ordering
from pathlib import Path
from typing import List


@dataclass
class TrackedPrompt:
    name: str
    repository_path: Path
    current_tag: str
    branches: List[str]


class ChangeNature(Enum):
    PromptTweak = 1
    InterfaceUpdate = 2
    ScopeChange = 3


@total_ordering
class PromptVersion:
    class IncrementerType(Enum):
        Major = 1
        Minor = 2
        Fix = 3

    def __init__(
        self, version: str, incrementer_type: IncrementerType = IncrementerType.Major
    ):
        self.incrementer_type = incrementer_type
        self.major, self.minor, self.fix = [int(x) for x in str(version).split(".")]

    def __version_id__(self) -> int:
        v = int(self.major) * 1e5 + int(self.minor) * 1e4 + int(self.fix) * 1e3
        return int(v)

    def __add__(self, x: int):
        if self.incrementer_type == PromptVersion.IncrementerType.Major:
            self.major += 1
        elif self.incrementer_type == PromptVersion.IncrementerType.Minor:
            self.minor += 1
        elif self.incrementer_type == PromptVersion.IncrementerType.Fix:
            self.fix += 1

        return self

    def __str__(self):
        return f"{self.major}.{self.minor}.{self.fix}"

    def __lt__(self, other):
        if not isinstance(other, PromptVersion):
            raise NotImplementedError()
        return self.__version_id__() < other.__version_id__()

    def __eq__(self, other):
        if not isinstance(other, PromptVersion):
            raise NotImplementedError()
        return self.__version_id__() == other.__version_id__()

    def __hash__(self):
        return hash((self.major, self.minor, self.fix))

    def set_incrementer(self, incremeter_type: IncrementerType):
        self.incrementer_type = incremeter_type



from abc import ABC, abstractmethod
from dataclasses import dataclass
from enum import Enum
import threading
from queue import Queue
from typing import Any, Callable, Dict, Optional


RegisterListenerFunction = Callable[[str, Callable[..., Any]], Any]
UnregisterListenerFunction = RegisterListenerFunction


class EventNotifier:
    """
    Helper class that notifies caller when an event is complete.

    The EventNotifier is usually returned by an external class
    (let us say Manager) that accepts registering listeners to
    its events. A common use is to create a context and check
    the value of the notifier.

    with Manager.get_notifier(event_name) as notifier:
        while True:
            if notifier.is_set():
                break
            do_something()
    """

    def __init__(
        self,
        event_name: str,
        register_fn: RegisterListenerFunction,
        unregister_fn: UnregisterListenerFunction,
    ):
        self.event_name = event_name
        self._event = threading.Event()
        self._register = register_fn
        self._unregister = unregister_fn

    def _callback(self, **kwargs):
        self._event.set()

    def __enter__(self):
        self._register(self.event_name, self._callback)
        return self._event

    def __exit__(self, exc_type, exc_value, traceback):
        self._unregister(self.event_name, self._callback)


class TaskListenerNotRegisteredError(Exception):
    pass


class Task(ABC):
    """
    Represent a task of TaskRunner.
    """

    class Event(Enum):
        Stop = "stop"
        Complete = "complete"

    def __init__(self):
        self._listeners = {}

    def _trigger_event(self, event: Event, **kwargs):
        if event not in self._listeners:
            return

        for callback in self._listeners[event]:
            callback(**kwargs)

    def _run(self, **kwargs):
        output = self.run(**kwargs)
        self._trigger_event(Task.Event.Complete, output=output)
        return output

    @abstractmethod
    def run(self, **kwargs):
        """
        The task logic.
        """
        ...

    def start(self, **kwargs):
        """
        Start executing the task logic.

        It triggers all listeners of Task.Event.Complete whenever
        is finished.
        """
        return self._run(**kwargs)

    def stop(self):
        """
        Trigger all listeners of event Task.Event.Stop.

        The logic contained in run is not aborted.
        It is up to the user to take this decision.

        # Run logic
        with self.get_event_notifier(Task.Event.Stop) as notifier:
            if notifier.is_set():
                take action
        """
        self._trigger_event(Task.Event.Stop)

    def register_listener(self, event: Event, callback):
        """
        Register a callback function to be called whenever the event is triggered.
        """
        if event not in self._listeners:
            self._listeners[event] = []
        self._listeners[event].append(callback)

    def unregister_listener(self, event: Event, callback):
        """
        Unregister a previously registered listener.

        Raises:
            TaskListenerNotRegisteredError if the listener to unregister is not found.
        """
        if event not in self._listeners:
            raise TaskListenerNotRegisteredError()

        filter_result = filter(lambda x: x is callback, self._listeners[event])
        self._listeners[event] = list(filter_result)

    def get_event_notifier(self, event: Event) -> EventNotifier:
        """
        Create EventNotifier for an event.

        The EventNotifier allows us to check if events such as Task.Event.Complete or
        Task.Event.Stop were triggered.
        """

        def _register(event_name: str, callback):
            self.register_listener(Task.Event(event_name), callback)

        def _unregister(event_name: str, callback):
            self.unregister_listener(Task.Event(event_name), callback)

        return EventNotifier(event.value, _register, _unregister)


class TaskNotRegisteredError(Exception):
    pass


@dataclass
class TaskInstruction:
    task_name: str
    task_input: Optional[Dict[str, Any]]


class TaskRunner:
    """
    Execute TaskInstruction added to its queue until emptness.

    TaskRunner executes tasks in the same order they are added in its internal
    queue via a TaskInstruction. The TaskInstruction contains the task name and
    its input variables. To execute a task, the latter needs to be priorly
    registered.

    The Task output is an optional TaskInstruction and it is added to the
    TaskRunner queue as soon it is completed. The added TaskInstruction is
    then executed in the next turn.

    The queue object is adapted to concurrent thread execution.
    """

    QuitInstruction = TaskInstruction("quit", None)

    def __init__(self):
        self._nodes: Dict[str, Optional[Task]] = {}
        self._callback: Dict[str, Any] = {}
        self._task_queue: Queue = Queue()
        self._current_task: Optional[str] = None

    def _register_task(self, task_name: str, task: Task, callback=None):
        self._nodes[task_name] = task
        self._callback[task_name] = callback

    def register(self, task_name: str, first_task: bool = False, callback=None):
        """
        Decorator to register a function as a task.
        """

        def decorator(func):
            class _Task(Task):
                def run(self, **kwargs):
                    return func(self, **kwargs)

            self._register_task(task_name, _Task(), callback)
            if first_task:
                self._task_queue = Queue()
                self._task_queue.put(TaskInstruction(task_name, None))

        return decorator

    def run(self):
        """
        Traverse the queue executing all tasks there present.
        """
        while not self._task_queue.empty():
            td = self._task_queue.get()
            if not td:
                continue

            if td == TaskRunner.QuitInstruction:
                break

            task = self._nodes[td.task_name]
            if not task:
                continue
            self._current_task = td.task_name

            if self._callback[td.task_name]:
                self._callback[td.task_name](td.task_input)

            input_dict = {}
            if td and td.task_input:
                input_dict = td.task_input

            self._task_queue.put(task.start(**input_dict))

    def next(self):
        """
        Stop the current task.
        """
        if not self._current_task:
            return
        task = self._nodes[self._current_task]
        if task:
            task.stop()

    def add(self, td: TaskInstruction):
        """
        Add a new TaskInstruction to the queue.
        """
        if td.task_name not in self._nodes:
            raise TaskNotRegisteredError()
        self._task_queue.put(td)

    def stop(self):
        """
        Stop all tasks.
        """
        self._task_queue.put(TaskRunner.QuitInstruction)

    def clear(self):
        """
        Clear task queue.
        """
        while not self._task_queue.empty():
            self._task_queue.get()



"""
Centralized logging configuration.
"""
import logging.config
from sys import stderr

LOGGING_CONFIG = {
    "version": 1,
    "disable_existing_loggers": False,
    "formatters": {
        "simple": {
            "format": "[%(asctime)s - %(name)s]-[%(levelname)s]: %(message)s",
        },
        "detailed": {
            "format": "[%(asctime)s - %(name)s - %(module)s::%(funcName)s]-[%(levelname)s]: %(message)s",
        },
        "info": {"format": "[%(levelname)s]: %(message)s"},
    },
    "handlers": {
        "stderr": {
            "level": "INFO",
            "class": "logging.StreamHandler",
            "formatter": "info",
            "stream": "ext://sys.stderr",
        },
        "file": {
            "level": "DEBUG",
            "class": "logging.FileHandler",
            "filename": "llm-assistant.log",
            "formatter": "detailed",
        },
        "console": {
            "level": "DEBUG",
            "class": "logging.StreamHandler",
            "formatter": "simple",
        },
    },
    "loggers": {
        "danoan.llm_assistant": {
            "level": "INFO",
            "handlers": ["stderr", "file"],
            "propagate": False,
        }
    },
}


def setup_logging():
    logging.config.dictConfig(LOGGING_CONFIG)



from danoan.llm_assistant.common.logging_config import setup_logging
from danoan.llm_assistant.runner.cli import utils as cli_utils
from danoan.llm_assistant.common import config as config


import argparse
import io
import json
import logging
import sys
from typing import List, Optional, TextIO

setup_logging()
logger = logging.getLogger(__name__)


def __run__(
    prompt_name: str,
    prompt_input: TextIO,
    from_text: Optional[str] = None,
    prompt_param: Optional[List[List[str]]] = None,
    *args,
    **kwargs,
):
    """
    Run a pre-registered prompt.
    """
    from danoan.llm_assistant.runner.cli.commands.run import run as M

    cli_utils.ensure_configuration_file_exists(logger)
    cli_utils.ensure_prompt_exists(prompt_name, logger)
    llma_config = config.get_configuration()

    prompt = config.get_prompt_configuration(prompt_name)
    ss = prompt_input
    if from_text:
        ss = io.StringIO()
        json.dump(json.loads(from_text), ss, ensure_ascii=False)
        ss.seek(0)

    response = M.run(llma_config.runner, prompt, ss, prompt_param)
    try:
        obj = json.loads(response.content)
        json.dump(obj, sys.stdout, indent=2, ensure_ascii=False)
    except Exception:
        wrapped = [response.content]
        json.dump(wrapped[0], sys.stdout, indent=2, ensure_ascii=False)


def extend_parser(subparser_action=None):
    command_name = "run"
    description = __run__.__doc__
    help = description.split(".")[0] if description else ""

    if subparser_action:
        parser = subparser_action.add_parser(
            command_name,
            help=help,
            description=description,
            formatter_class=argparse.RawDescriptionHelpFormatter,
        )
    else:
        parser = argparse.ArgumentParser(
            command_name,
            description=description,
            formatter_class=argparse.RawDescriptionHelpFormatter,
        )

    parser.add_argument("prompt_name", type=str, help="Name of the prompt to run")
    parser.add_argument(
        "--prompt-param",
        "--p",
        nargs=2,
        action="append",
        type=str,
        help="Key value pair specifing one parameter accepted by the prompt",
    )

    meg = parser.add_mutually_exclusive_group()
    meg.add_argument(
        "--from-text",
        type=str,
        help="Prompt parameters passed as a JSON object encoded as string",
    )
    meg.add_argument(
        "prompt_input",
        nargs="?",
        type=argparse.FileType("r"),
        default=sys.stdin,
        help="Prompt input in json format",
    )

    parser.set_defaults(func=__run__, subcommand_help=parser.print_help)

    return parser



from danoan.llm_assistant.common.logging_config import setup_logging
from danoan.llm_assistant.runner.cli import utils as cli_utils
from danoan.llm_assistant.common.config import LLM_ASSISTANT_ENV_VARIABLE

import argparse
import logging

setup_logging()
logger = logging.getLogger(__name__)


def __init_llm_assistant__(
    reset: bool = False, use_env_var: bool = False, *args, **kwargs
):
    """
    Initialize llm-assistant configuration.

    It creates the configuration file in the location pointed
    by the environment variable LLM_ASSISTANT_CONFIGURATION_FOLDER.
    """
    from danoan.llm_assistant.runner.cli.commands.setup.setup_commands import init as M

    cli_utils.ensure_environment_variable_is_defined(logger)
    M.init(reset=reset, use_env_var=use_env_var)


def extend_parser(subparser_action=None):
    command_name = "init"
    description = __init_llm_assistant__.__doc__
    help = description.split(".")[0] if description else ""

    if subparser_action:
        parser = subparser_action.add_parser(
            command_name,
            help=help,
            description=description,
            formatter_class=argparse.RawDescriptionHelpFormatter,
        )
    else:
        parser = argparse.ArgumentParser(
            command_name,
            description=description,
            formatter_class=argparse.RawDescriptionHelpFormatter,
        )

    parser.add_argument(
        "--env",
        dest="use_env_var",
        action="store_true",
        help=f"Creates configuration file in the directory pointed by the environment variable {LLM_ASSISTANT_ENV_VARIABLE}",
    )
    parser.add_argument(
        "--force",
        dest="reset",
        action="store_true",
        help="Rewrites the configuration file",
    )
    parser.set_defaults(func=__init_llm_assistant__, subcommand_help=parser.print_help)

    return parser



from danoan.llm_assistant.common.logging_config import setup_logging
from danoan.llm_assistant.common import config
from danoan.llm_assistant.prompt.core import api

from git.remote import RemoteProgress
import logging
from typing import Dict

setup_logging()
logger = logging.getLogger(__name__)


def __git_event_processor__(op_code, cur_count, max_count=None, message=""):
    d = " "
    if not op_code & RemoteProgress.RECEIVING:
        return

    if max_count:
        percent = (cur_count / max_count) * 100
        logger.info(f"{d*4}Progress: {percent:.2f}% - {message}")


def __get_sync_events_processor__():
    params: Dict = {}
    action = None

    def inner(event, name, value):
        nonlocal params
        nonlocal action
        d = " "
        if event == "sync_config":
            if name is None:
                logger.info(
                    "Starting synchronization of prompt collection folder with configuration file"
                )
            else:
                logger.info(f"{d*2}{name}: {value}")
        elif event == "fetch":
            logger.info(f"{d*4}Fetching repository: {value}")
        elif event == "checkout":
            logger.info(f"{d*4}Checkout version: {value}")
        elif event == "synced":
            logger.info(f"{d*4}Already synced")
        elif event == "sync_prompt_collection_folder":
            if name is None:
                logger.info(
                    "Starting synchronization of configuration file with prompt collection folder"
                )
            else:
                logger.info(f"{d*2}{name}: {value}")
        elif event == "not_tracked":
            logger.info(f"{d*4}Prompt not tracked. Registering version: {value}")
        elif event == "not_prompt_repository":
            logger.info(f"{d*4}This is not a prompt repository. Skipping it.")
        elif event == "git":
            action = __git_event_processor__
        elif event == "begin":
            params = {}
        elif event == "item":
            params[name] = value
        elif event == "end" and action:
            action(**params)

    return inner


def sync():
    llma_config = config.get_configuration()
    api.sync(llma_config.prompt, __get_sync_events_processor__())



from abc import ABC, abstractmethod


class CLIDrawer(ABC):
    @abstractmethod
    def print_error(self, **kwargs):
        pass

    @abstractmethod
    def print_panel(self, **kwargs):
        pass

    @abstractmethod
    def print_list(self, **kwargs):
        pass

    @abstractmethod
    def prompt(self, **kwargs):
        pass



from danoan.llm_assistant.common.logging_config import setup_logging
from danoan.llm_assistant.common import model, utils
from danoan.llm_assistant.runner.core import api

import json
import logging
from typing import List, Optional, TextIO

setup_logging()
logger = logging.getLogger(__name__)


def run(
    runner_configuration: model.RunnerConfiguration,
    prompt: model.PromptConfiguration,
    prompt_input: TextIO,
    prompt_param: Optional[List[List[str]]],
):
    api.LLMAssistant().setup(
        utils.generate_absolute_runner_config(runner_configuration)
    )

    input_obj = json.load(prompt_input)
    if prompt_param:
        for key, value in prompt_param:
            input_obj[key] = value

    return api.custom(prompt, **input_obj)



from copy import deepcopy
from danoan.llm_assistant.common import model
from danoan.llm_assistant.runner.cli import utils as cli_utils
from danoan.llm_assistant.runner.cli.commands.session.cli_drawer import CLIDrawer
from danoan.llm_assistant.runner.cli.commands.session.task_runner import (
    Task,
    TaskInstruction,
)
from danoan.llm_assistant.runner.core import api

from dataclasses import dataclass
from enum import Enum
from pathlib import Path
import toml
from typing import Any, Callable, Dict, Optional


#############################################
# Local model
#############################################


class PromptInterruptedError(Exception):
    pass


@dataclass
class PromptConfigurationWithLocation:
    location: Path
    prompt_config: model.PromptConfiguration


#############################################
# Helper functions
#############################################


ValidationFunction = Callable[[Any], bool]
PromptFunction = Callable[[], str]


def _retry_prompt_until_valid_input(
    prompt_fn: PromptFunction,
    validation_fn: ValidationFunction,
) -> str:
    """
    Ask for user prompt until a condition is valid.

    Prompt blocks the execution until the user enter with some
    input. It is possible to take back the control of the execution
    by checking the task instance that originated the prompt execution.

    Args:
        prompt_fun: Function to collect user input.
        validation_fn: Function that validates user input.
    Raises:
        PromptInterruptedError if prompt is interrupted.
        ValueError if entered value has None value.
    """

    entered_value = None

    while not validation_fn(entered_value):
        try:
            entered_value = prompt_fn()
        except PromptInterruptedError as ex:
            raise ex

    if not entered_value:
        raise ValueError()

    return entered_value


def _validate_int(
    value: Optional[str],
    floor: int,
    ceiling: int,
) -> bool:
    """
    Checks if value is an integer and within the range defined by floor and ceiling.

    Args:
        value: String value to check if it represents an integer.
        floor: Lower bound for the integer value.
        ceiling: Upper bound for the integer value.
    Raises:
        ValueError if string value cannot be converted to integer.
        IndexError if value is out of the [floor,ceiling] range.
    """
    if not value:
        return False

    v = int(value)
    if v >= floor and v <= ceiling:
        return True
    else:
        raise IndexError()

    return False


def _retry_prompt_until_valid_int(
    message: str, floor: int, ceiling: int, cliDrawer: CLIDrawer, task: Task
):
    """
    Prompt until a valid integer is entered.

    Args:
        value: String value to check if it represents an integer.
        floor: Lower bound for the integer value.
        ceiling: Upper bound for the integer value.
        cliDrawer: Interface to collect prompt input and print eventual error messages.
        task: Task from which the prompt was requested.
    """

    def validator(value) -> bool:
        try:
            return _validate_int(value, floor, ceiling)
        except IndexError:
            cliDrawer.print_error(message="Index out of bound")
        except ValueError:
            cliDrawer.print_error(message=f"Invalid integer value: `{value}`")

        return False

    def prompt_fn():
        return cliDrawer.prompt(message=message, task=task)

    return _retry_prompt_until_valid_input(prompt_fn, validator)


class TaskName(Enum):
    NoPromptSelected = ("NoPromptSelected",)
    PromptSelected = ("PromptSelected",)
    NewInstance = ("NewInstance",)
    LoadInstance = ("LoadInstance",)
    StartChat = ("StartChat",)
    ContinueChat = ("ContinueChat",)

    def __str__(self):
        return self.value


def register_tasks(
    cliDrawer: CLIDrawer,
    prompt_repository: Path,
    register_function,
):
    """
    Register the session runner tasks.

    NoPromptSelected -> PromptSelected                   -----> StartChat -> ContinueChat
                            |                            |
                            ------> NewInstance -------- |
                            |                            |
                            ------> LoadInstance ------- |
    """

    @register_function(TaskName.NoPromptSelected, first_task=True)
    def _no_prompt_selected(task: Task) -> Optional[TaskInstruction]:
        list_prompt_config = []
        for prompt_config_filepath in prompt_repository.rglob("*config.toml"):
            _, prompt_config = cli_utils.is_a_prompt_config_file(prompt_config_filepath)
            if prompt_config:
                list_prompt_config.append(
                    PromptConfigurationWithLocation(
                        prompt_config_filepath, prompt_config
                    )
                )

        if len(list_prompt_config) == 0:
            cliDrawer.print_error(
                message=f"No prompt available in the repository: {prompt_repository}"
            )
            return TaskInstruction("NoPromptSelected", None)

        cliDrawer.print_panel(message="Select prompt")
        cliDrawer.print_list(
            list_elements=[e.prompt_config.name for e in list_prompt_config]
        )

        try:
            entered_str = _retry_prompt_until_valid_int(
                "Prompt index: ",
                1,
                len(list_prompt_config),
                cliDrawer,
                task,
            )
            prompt_index = int(entered_str)
        except PromptInterruptedError:
            return None

        return TaskInstruction(
            TaskName.PromptSelected,
            {"prompt_config": list_prompt_config[prompt_index - 1]},
        )

    @register_function(TaskName.PromptSelected)
    def _prompt_selected(
        task: Task, prompt_config: PromptConfigurationWithLocation
    ) -> Optional[TaskInstruction]:
        cliDrawer.print_panel(message="Choose an option")
        cliDrawer.print_list(list_elements=["New instance", "Load instance"])

        try:
            entered_str = _retry_prompt_until_valid_int(
                "Option index: ",
                1,
                2,
                cliDrawer,
                task,
            )
            option_index = int(entered_str)
        except PromptInterruptedError:
            return None

        next_state = None
        if option_index == 1:
            next_state = TaskName.NewInstance
        elif option_index == 2:
            next_state = TaskName.LoadInstance
        elif option_index is None:
            next_state = TaskName.PromptSelected
        else:
            return None

        return TaskInstruction(next_state, {"prompt_config": prompt_config})

    @register_function(TaskName.NewInstance)
    def _new_instance(
        task: Task, prompt_config: PromptConfigurationWithLocation
    ) -> Optional[TaskInstruction]:
        instances_folder = (
            prompt_repository / "instances" / prompt_config.location.parent.name
        )
        instances_folder.mkdir(parents=True, exist_ok=True)
        instance_filepaths = list(instances_folder.iterdir())
        list_instance_names = [f.stem for f in instance_filepaths]

        def validate_instance_name(value: str) -> bool:
            if not value:
                return False

            is_valid = value not in list_instance_names
            if not is_valid:
                cliDrawer.print_error(
                    callback=None,
                    message="This instance name exists already. Choose another one.",
                )

            return is_valid

        instance_name = None
        try:
            instance_name = _retry_prompt_until_valid_input(
                lambda: cliDrawer.prompt(
                    message="Enter the instance name: ", task=task
                ),
                validate_instance_name,
            )
        except PromptInterruptedError:
            return None

        instance_input = {}
        for input_variable in api.get_prompt_input_variables(
            prompt_config.prompt_config
        ):
            instance_input_value = None
            try:
                instance_input_value = cliDrawer.prompt(
                    message=f"{input_variable} (empty to skip): ", task=task
                )
            except PromptInterruptedError:
                return None

            print("input;", instance_input_value)
            instance_input_value = instance_input_value.strip()
            if len(instance_input_value) > 0:
                instance_input[input_variable] = instance_input_value

        assignments_str = "::".join(
            [f"{name}={value}" for name, value in instance_input.items()]
        )
        print("HEYHEYHEY--", assignments_str, "--")

        instance_filepath = instances_folder / f"{instance_name}.toml"

        instance: Dict[str, Any] = {"name": instance_name}
        variables: Dict[str, Any] = {}
        for e in assignments_str.split("::"):
            k, v = e.split("=")
            variables[k] = v
        instance["variables"] = variables

        with open(instance_filepath, "w") as f:
            toml.dump(instance, f)

        return TaskInstruction(
            TaskName.StartChat,
            {"prompt_config": prompt_config.prompt_config, "instance": instance},
        )

    @register_function(TaskName.LoadInstance)
    def _load_instance(
        task: Task, prompt_config: PromptConfigurationWithLocation
    ) -> Optional[TaskInstruction]:
        instances_folder = (
            prompt_repository / "instances" / prompt_config.location.parent.name
        )

        if not instances_folder.exists():
            cliDrawer.print_error(message=f"No instance available for this prompt")
            return TaskInstruction(
                TaskName.PromptSelected, {"prompt_config": prompt_config}
            )

        instance_filepaths = list(instances_folder.iterdir())
        instance_names = [f.stem for f in instance_filepaths]

        cliDrawer.print_panel(callback=None, message="Select instance")
        cliDrawer.print_list(callback=None, list_elements=instance_names)

        try:
            entered_str = _retry_prompt_until_valid_int(
                "Select instance index: ",
                1,
                len(instance_names),
                cliDrawer,
                task,
            )
            instance_index = int(entered_str)
        except PromptInterruptedError:
            return None

        instance_filepath = instance_filepaths[instance_index - 1]

        with open(instance_filepath, "r") as f:
            instance = toml.load(f)

        return TaskInstruction(
            TaskName.StartChat,
            {"prompt_config": prompt_config.prompt_config, "instance": instance},
        )

    @register_function(TaskName.StartChat)
    def _start_chat(
        task: Task, prompt_config: model.PromptConfiguration, instance: Dict[str, Any]
    ) -> Optional[TaskInstruction]:
        message = f"Prompt: {prompt_config.name}\nInstance:{instance['name']}"

        cliDrawer.print_panel(
            callback=None, message=message, color="pink3", title="New Chat"
        )

        return TaskInstruction(
            TaskName.ContinueChat,
            {"prompt_config": prompt_config, "instance": instance},
        )

    @register_function(TaskName.ContinueChat)
    def _continue_chat(
        task: Task, prompt_config: model.PromptConfiguration, instance: Dict[str, Any]
    ) -> Optional[TaskInstruction]:
        chat_title = f"{prompt_config.name}::{instance['name']}"

        message_lines = []
        try:
            while True:
                # TODO: Create a variation of prompt that accept message
                # with several lines
                line = cliDrawer.prompt(task=task, message="Enter message: ")
                if line == "$$":
                    break
                message_lines.append(line)
        except PromptInterruptedError:
            return None

        message = "\n".join(message_lines)

        cliDrawer.print_panel(message=message, color="cyan", title=chat_title)

        response = api.custom(prompt_config, message=message, **instance["variables"])
        cliDrawer.print_panel(
            message=response.content, color="orange1", title=chat_title
        )

        return TaskInstruction(
            TaskName.ContinueChat,
            {"prompt_config": prompt_config, "instance": instance},
        )



"""
prompt-manager utility tools.
"""

from danoan.llm_assistant.common.logging_config import setup_logging

from functools import lru_cache
import git
import logging
from pathlib import Path
from typing import Any, List

setup_logging()
logger = logging.getLogger(__name__)

###################################
# GIT Helpers
###################################


def get_commit_tags(repository_folder: Path, commit_hash: str) -> List[str]:
    """
    Get all tags of a commit.
    """
    repo = git.Repo(repository_folder)
    commit = repo.commit(commit_hash)

    return [tag for tag in repo.tags if tag.commit == commit]


@lru_cache
def get_most_recent_tags_before_commit(
    repository_folder: Path, commit_hash: str
) -> List[git.Tag]:
    """
    Get a list of the most recent tags prior to a commit.
    """
    repository = git.Repo(repository_folder)

    for p in repository.iter_commits(commit_hash):
        tags = get_commit_tags(repository_folder, p.hexsha)
        if len(tags) > 0:
            return tags
    return []


@lru_cache
def get_all_commits_in_between(
    repository_folder: Path, most_recent_hash: str, most_ancient_hash: str
) -> List[str]:
    """
    Return a list of commit hashes in between two commit hashes (terminals included).
    """
    repository = git.Repo(repository_folder)
    commits = []
    for p in repository.iter_commits(most_recent_hash):
        commits.append(p.hexsha)
        if p.hexsha == most_ancient_hash:
            break
    return commits


@lru_cache
def get_non_versioned_commits(repository_folder: Path) -> List[str]:
    """
    Get the first sequence of non-versioned commit hashes.

    Start from the HEAD commit up to the first tagged commit (included).
    """
    repository = git.Repo(repository_folder)
    most_recent_tags = get_most_recent_tags_before_commit(
        repository_folder, repository.commit()
    )
    if len(most_recent_tags) == 0:
        return []
    tag = most_recent_tags[-1]
    return get_all_commits_in_between(
        repository_folder, repository.commit(), tag.commit
    )


def get_commit(repository_folder: Path, commit_hash: str) -> git.Commit:
    repository = git.Repo(repository_folder)
    return repository.commit(commit_hash)


def push_new_version(repository_folder: Path, version: str):
    repository = git.Repo(repository_folder)
    tag = repository.create_tag(
        version, repository.commit(), message=f"Release version {version}"
    )
    repository.remote().push(tag)


def get_staging_area(repository_folder: Path) -> List[Any]:
    repository = git.Repo(repository_folder)
    return repository.commit().diff()


def get_versions(repository_path: Path) -> List[str]:
    repo = git.Repo(repository_path)
    return [t.name[1:] for t in repo.tags]


def get_branches_names(repository_path: Path) -> List[str]:
    repo = git.Repo(repository_path)
    return ["/".join(x.name.split("/")[1:]) for x in repo.remote().refs]
